{createStorageAdapter, TestStorageAdapter, AsyncTestStorageAdapter} = if typeof require isnt 'undefined' then require '../model_helper' else window
helpers = if typeof require is 'undefined' then window.viewHelpers else require '../../view/view_helper'

QUnit.module "Batman.Model belongsTo Associations"
  setup: ->
    namespace = @namespace = this
    class @Store extends Batman.Model
      @encode 'id', 'name'

    @storeAdapter = createStorageAdapter @Store, AsyncTestStorageAdapter,
      'stores1':
        name: "Store One"
        id: 1
      'stores2':
        name: "Store Two"
        id: 2
        product:
          id:3
          name:"JSON Product"
      'stores3':
        name: "Store Three"
        id: 3

    class @Product extends Batman.Model
      @encode 'id', 'name'
      @belongsTo 'store', namespace: namespace
    @productAdapter = createStorageAdapter @Product, AsyncTestStorageAdapter,
      'products1':
        name: "Product One"
        id: 1
        store_id: 1
      'products4':
        name: "Product One"
        id: 1
        store:
          name: "Store Three",
          id: 3

asyncTest "belongsTo yields the related model when toJSON is called", 1, ->
  @Product.find 1, (err, product) =>
    store = product.get('store')
    storeJSON = store.toJSON()
    # store will encode its product
    delete storeJSON.product

    deepEqual storeJSON, @storeAdapter.storage["stores1"]
    QUnit.start()

asyncTest "belongsTo associations are loaded via ID", 1, ->
  @Product.find 1, (err, product) =>
    store = product.get 'store'
    equal store.get('id'), 1
    QUnit.start()

asyncTest "belongsTo associations are not loaded when autoload is off", 1, ->
  class Product extends Batman.Model
    @encode 'id', 'name'
    @belongsTo 'store', {namespace: @namespace, autoload: false}

  productAdapter = createStorageAdapter Product, AsyncTestStorageAdapter,
    'products1': {name: "Product One", id: 1, store_id: 1}

  Product.find 1, (err, product) =>
    store = product.get 'store'
    equal (typeof store), 'undefined'
    QUnit.start()

asyncTest "belongsTo associations are saved", 5, ->
  store = new @Store name: 'Zellers'
  product = new @Product name: 'Gizmo'
  product.set 'store', store

  productSaveSpy = spyOn product, 'save'
  product.save (err, record) =>
    equal productSaveSpy.callCount, 1
    equal record.get('store_id'), store.id
    storedJSON = @productAdapter.storage["products#{record.id}"]
    deepEqual storedJSON, product.toJSON()

    store = record.get('store')
    equal storedJSON.store_id, undefined

    @Product.find record.get('id'), (err, product2) ->
      deepEqual product2.toJSON(), storedJSON
      QUnit.start()

asyncTest "belongsTo parent models are added to the identity map", 1, ->
  @Product.find 4, (err, product) =>
    equal @Store.get('loaded').length, 1
    QUnit.start()

asyncTest "belongsTo parent models are passed through the identity map", 2, ->
  @Store.find 3, (err, store) =>
    throw err if err
    @Product.find 4, (err, product) =>
      equal @Store.get('loaded').length, 1
      equal product.get('store'), store
      QUnit.start()

asyncTest "belongsTo associations render", 1, ->
  @Product.find 1, (err, product) ->
    source = '<span data-bind="product.store.name"></span>'
    context = Batman(product: product)
    helpers.render source, context, (node) =>
      equal node[0].innerHTML, 'Store One'
      QUnit.start()

asyncTest "belongsTo supports inline saving", 1, ->
  namespace = this
  class @InlineProduct extends Batman.Model
    @encode 'name'
    @belongsTo 'store', namespace: namespace, saveInline: true
  storageAdapter = createStorageAdapter @InlineProduct, AsyncTestStorageAdapter

  product = new @InlineProduct name: "Inline Product"
  store = new @Store name: "Inline Store"
  product.set 'store', store

  product.save (err, record) =>
    deepEqual storageAdapter.storage["inline_products#{record.get('id')}"],
      name: "Inline Product"
      store: {name: "Inline Store"}
    QUnit.start()

QUnit.module "Batman.Model belongsTo Associations with inverseOf to a hasMany"
  setup: ->
    namespace = @namespace = this

    class @Order extends Batman.Model
      @encode 'id', 'name'
      @belongsTo 'customer', {namespace: namespace, inverseOf: 'orders'}

    @orderAdapter = createStorageAdapter @Order, AsyncTestStorageAdapter,
      'orders1':
        name: "Order One"
        id: 1
        customer:
          name: "Customer One"
          id: 1
      'orders2':
        name: "Order Two"
        id: 1
        customer:
          name: "Customer One"
          id: 1

    class @Customer extends Batman.Model
      @encode 'id', 'name'
      @hasMany 'orders', namespace: namespace

    @customerAdapter = createStorageAdapter @Customer, AsyncTestStorageAdapter,
      'customers1':
        name: "Customer One"
        id: 1

asyncTest "belongsTo sets the foreign key on itsself so the parent relation SetIndex adds it, if the parent hasn't been loaded", 1, ->
  @Order.find 1, (err, order) =>
    throw err if err
    customer = order.get('customer')
    delay ->
      ok customer.get('orders').has(order)

asyncTest "belongsTo sets the foreign key on itself so the parent relation SetIndex adds it, if the parent has already been loaded", 1, ->
  @Customer.find 1, (err, customer) =>
    throw err if err
    @Order.find 1, (err, order) =>
      throw err if err
      customer = order.get('customer')
      delay ->
        ok customer.get('orders').has(order)

asyncTest "belongsTo sets the foreign key foreign key on itsself such that many loads are picked up by the parent", 3, ->
  @Customer.find 1, (err, customer) =>
    throw err if err
    @Order.find 1, (err, order) =>
      throw err if err
      equal customer.get('orders').length, 1
      @Order.find 2, (err, order) =>
        throw err if err
        equal customer.get('orders').length, 2
        equal @Customer.get('loaded').length, 1, 'Only one parent record should be created'
        QUnit.start()


QUnit.module "Batman.Model belongsTo Associations with inverseOf to a hasOne"
  setup: ->
    namespace = @namespace = this

    class @Order extends Batman.Model
      @encode 'id', 'name'
      @belongsTo 'customer', {namespace: namespace, inverseOf: 'order'}

    @orderAdapter = createStorageAdapter @Order, AsyncTestStorageAdapter,
      'orders1':
        name: "Order One"
        id: 1
        customer:
          name: "Customer One"
          id: 1

    class @Customer extends Batman.Model
      @encode 'id', 'name'
      @hasOne 'order', namespace: namespace

    @customerAdapter = createStorageAdapter @Customer, AsyncTestStorageAdapter,
      'customers1':
        name: "Customer One"
        id: 1

asyncTest "belongsTo sets the inverse relation if the parent hasn't been loaded", 1, ->
  @Order.find 1, (err, order) =>
    throw err if err
    customer = order.get('customer')
    equal customer.get('order'), order
    QUnit.start()

asyncTest "belongsTo sets the inverse relation if the parent has already been loaded", 1, ->
  @Customer.find 1, (err, customer) =>
    throw err if err
    @Order.find 1, (err, order) =>
      throw err if err
      customer = order.get('customer')
      equal customer.get('order'), order
      QUnit.start()
